/*
 * (C) Copyright IBM Corp. 2003    All Rights Reserved.
 * mdctrl.c ($Revision: 1.5 $)
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corp.
 *
 * Author(s): S. Bader <shbader@de.ibm.com>
 *
 * This file is part of MDMPD.
 *
 * MDMPD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MDMPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MDMPD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>

#include <linux/kdev_t.h>
#include <linux/raid/md_u.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include <mdstat.h>
#include <mdctrl.h>

#define ERR(x, ...)	fprintf(stderr, "%s @ %d: " x, __FUNCTION__ , \
				__LINE__ , ##__VA_ARGS__)
#if DEBUG > 4
#define DBG(x, ...)	fprintf(stderr, "%s @ %d: " x, __FUNCTION__ , \
				__LINE__ , ##__VA_ARGS__)
#else
#define DBG(x, ...)
#endif

static char *	revision		= "mdctrl.c ($Revision: 1.5 $)";
static char 	node_directory[]	= "/var/run/mdmpd";
static char	path[80];

static char *print_number(char *s, long number) {
	static char	n[20];
	static int	i;

	i = 0;
	while(number > 0) {
		n[i] = '0' + (number % 10);
		number /= 10;
		i++;
	}
	if(i == 0) {
		*s = '0';
		s++;
	} else {
		while(i > 0) {
			*s = n[i-1];
			s++;
			i--;
		}
	}
	*s = '\0';

	return s;
}

static char *print_string(char *t, char *s) {
	while(*s != '\0')
		*t++ = *s++;
	*t = '\0';

	return t;
}

void print_path(char *s, int md_nr, int disk_nr) {
	s = print_string(s, node_directory);
	s = print_string(s, "/md");
	s = print_number(s, md_nr);
	if(disk_nr >= 0) {
		s = print_string(s, "_disk");
		s = print_number(s, disk_nr);
	}
}

int make_directory(char *name) {
	struct stat	st;
	int		rc;

	(void) revision;

	if((rc = stat(name, &st)) < 0 && errno != ENOENT) {
		return rc;
	}

	if(rc == 0) {
		if(!S_ISDIR(st.st_mode)) {
			ERR("Exists but not as directory\n");
			errno = ENOTDIR;
			return -1;
		}
		return 0;
	}

	DBG("mkdir(%s)\n", name);
	return mkdir(name, S_IRWXU);
}

int make_node_directory(void) {
	static char		path[80];
	int			rc;
	int			i;

	rc = 0;
	for(i = 0; node_directory[i] != '\0'; i++) {
		if(node_directory[i] == '/' && i > 0) {
			path[i] = '\0';
			if((rc = make_directory(path)) < 0)
				break;
		}
		path[i] = node_directory[i];
	}
	path[i] = '\0';

	if(rc == 0)
		rc = make_directory(path);

	return rc;
}

void md_get_array(struct mds_array *mda) {
	DBG("md%d: new array\n", mda->nr);
	print_path(path, mda->nr, -1);
	mknod(path, S_IFBLK|S_IRUSR|S_IWUSR, MKDEV(9, mda->nr));
}

void md_put_array(struct mds_array *mda) {
	DBG("md%d: removed array\n", mda->nr);
	print_path(path, mda->nr, -1);
	unlink(path);
}

void md_get_disk(struct mds_array *mda, int disk_nr) {
	DBG("md%d: new disk%d\n", mda->nr, disk_nr);
	print_path(path, mda->nr, disk_nr);
	mknod(
		path,
		S_IFBLK|S_IRUSR|S_IWUSR,
		MKDEV(
			mda->disk[disk_nr].major,
			mda->disk[disk_nr].minor
		)
	);
}

void md_put_disk(struct mds_array *mda, int disk_nr) {
	DBG("md%d: removed disk%d\n", mda->nr, disk_nr);
	print_path(path, mda->nr, disk_nr);
	unlink(path);
}

int md_reactivate_disk(struct mds_array *mda, int disk_nr) {
	static char		path[80];
	int			fd;
	int			rc;
	long			dev;

	DBG("md_reactivate_disk() called\n");
	print_path(path, mda->nr, -1);
	DBG("Open %s\n", path);
	if((fd = open(path, O_RDONLY)) < 0)
		return fd;

	dev = MKDEV(mda->disk[disk_nr].major, mda->disk[disk_nr].minor);
	DBG("Hot remove md%d disk%d\n", mda->nr, disk_nr);
	if((rc = ioctl(fd, HOT_REMOVE_DISK, dev)) == 0) {
		DBG("Hot add md%d disk%d\n", mda->nr, disk_nr);
		rc = ioctl(fd, HOT_ADD_DISK, dev);
	}

	close(fd);
	
	return rc;
}

int md_poll_disk(struct mds_array *mda, int disk_nr) {
	static char		buffer[10];
	static char		path[80];
	int			fd;
	ssize_t			rc;

	print_path(path, mda->nr, disk_nr);
	DBG("Probing %s\n", path);
	if((fd = open(path, O_RDONLY)) < 0) {
		DBG("Open failed\n");
		return fd;
	}

	do {
		rc = read(fd, buffer, sizeof(buffer));
	} while (rc < 0 && errno == EINTR);

	DBG("Probe returned %ld (errno=%d)\n", rc, errno);

	close(fd);

	if(rc > 0) {
		DBG("Calling reactivate_disk\n");
		rc = md_reactivate_disk(mda, disk_nr);
	} else
		rc = -1;

	return rc;
}

